package me.mrletsplay.videobase.library;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.stream.Collectors;

import me.mrletsplay.mrcore.json.JSONObject;
import me.mrletsplay.mrcore.json.JSONParseException;
import me.mrletsplay.mrcore.json.converter.JSONConverter;
import me.mrletsplay.videobase.VideoBase;
import me.mrletsplay.videobase.util.Hash;

public class Library {

	public static final String
		METADATA_NAME = "meta.json",
		VIDEO_METADATA_SUFFIX = ".meta.json";

	private Path path;
	private List<Video> videos;

	public Library(Path path, List<Video> videos) {
		Objects.requireNonNull(path, "path");
		Objects.requireNonNull(videos, "videos");

		this.path = path;
		this.videos = videos;
	}

	public Path getPath() {
		return path;
	}

	public List<Video> getVideos() {
		return videos;
	}

	public Video findVideoById(String id) {
		return videos.stream()
			.filter(v -> v.getId().equals(id))
			.findFirst().orElse(null);
	}

	/**
	 * Finds the videos for the given series and/or author sorted by their index
	 * @param series The series or {@code null} to not filter by series
	 * @param author The author or {@code null} to not filter by author
	 * @return The videos matching the given criteria
	 */
	public List<Video> findVideosBy(String series, String author) { // TODO: Take VideoMetadata to allow any parameter from meta
		return videos.stream()
			.filter(v -> (series == null || series.equals(v.getMetadata().getSeries()))
				&& (author == null || author.equals(v.getMetadata().getAuthor())))
			.sorted(Comparator.comparing(v -> v.getMetadata().getIndex()))
			.collect(Collectors.toList());
	}

	public List<String> getSeries() {
		return videos.stream()
			.map(v -> v.getMetadata().getSeries())
			.distinct()
			.collect(Collectors.toList());
	}

	public List<String> getAuthors() {
		return videos.stream()
			.map(v -> v.getMetadata().getAuthor())
			.distinct()
			.collect(Collectors.toList());
	}

	public void updateMetadata(Video video, VideoMetadata metadata) {
		VideoMetadata oldMetadata = video.getMetadata();
		video.setMetadata(VideoMetadata.inherit(oldMetadata, metadata));
		// TODO: save metadata
	}

	public boolean updateMetadata(String videoId, VideoMetadata metadata) {
		Video video = findVideoById(videoId);
		if(video == null) return false;
		updateMetadata(video, metadata);
		return true;
	}

	public static Library load(Path path) {
		List<Video> videos = new ArrayList<>();
		load(path, path, videos, VideoMetadata.DEFAULT_METADATA);
		return new Library(path, videos);
	}

	private static void load(Path rootPath, Path path, List<Video> videos, VideoMetadata parentDefaultMetadata) {
		if(!Files.isDirectory(path)) return;

		LibraryMetadata libraryMeta = null;
		VideoMetadata defaultMetadata = parentDefaultMetadata;

		Path metaPath = path.resolve("meta.json");
		if(Files.isRegularFile(metaPath)) {
			try {
				JSONObject obj = new JSONObject(Files.readString(metaPath));
				libraryMeta = JSONConverter.decodeObject(obj, LibraryMetadata.class);
				defaultMetadata = VideoMetadata.inherit(defaultMetadata, libraryMeta.getDefault());
			}catch(IOException | JSONParseException | ClassCastException e) {
				VideoBase.LOGGER.warn("Failed to parse metadata at " + path, e);
			}
		}

		try {
			for(Path subPath : Files.walk(path, 1)
					.filter(p -> path.equals(p.getParent()))
					.sorted(Comparator.comparing(p -> p.getFileName().toString()))
					.collect(Collectors.toList())) {
				String fileName = subPath.getFileName().toString();
				String fileNameNoExt = fileName;
				if(fileNameNoExt.contains(".")) {
					fileNameNoExt = fileNameNoExt.substring(0, fileNameNoExt.lastIndexOf('.'));
				}

				if(Files.isDirectory(subPath)) {
					load(rootPath, subPath, videos, defaultMetadata);
					continue;
				}

				if(fileName.equals(METADATA_NAME) || fileName.endsWith(VIDEO_METADATA_SUFFIX)) {
					continue;
				}

				VideoMetadata inferredMetadata = new VideoMetadata(new JSONObject(Map.of(
					VideoMetadata.FIELD_TITLE, fileNameNoExt,
					VideoMetadata.FIELD_SERIES, path.getFileName().toString()
				)));

				VideoMetadata videoMeta = VideoMetadata.inherit(defaultMetadata, inferredMetadata);

				Path videoMetaPath = path.resolve(subPath.getFileName().toString() + VIDEO_METADATA_SUFFIX);
				if(Files.isRegularFile(videoMetaPath)) {
					try {
						JSONObject obj = new JSONObject(Files.readString(videoMetaPath));
						videoMeta = VideoMetadata.inherit(videoMeta, JSONConverter.decodeObject(obj, VideoMetadata.class));
					}catch(IOException | JSONParseException | ClassCastException e) {
						VideoBase.LOGGER.warn("Failed to parse metadata at " + path, e);
					}
				}else if(libraryMeta != null && libraryMeta.getOverrides().containsKey(fileName)) {
					videoMeta = VideoMetadata.inherit(videoMeta, libraryMeta.getOverrides().get(fileName));
				}

				String id = Hash.hash(rootPath.relativize(subPath).toString());
				videos.add(new Video(subPath, id, videoMeta == null ? VideoMetadata.DEFAULT_METADATA : videoMeta));
			}
		}catch(IOException e) {
			VideoBase.LOGGER.warn("Failed to load folder at " + path, e);
		}
	}

}
